// Automatically generated by instruction-table.oak

char *argless_instr_name[] = {
  "NOOP",			/* 0 */
  "PLUS",
  "NEGATE",
  "EQ?",
  "NOT",
  "TIMES",
  "LOAD-GLO",
  "DIV",
  "=0?",
  "GET-TAG",
  "GET-DATA",			/* 10 */
  "CRUNCH",
  "GETC",
  "PUTC",
  "CONTENTS",
  "SET-CONTENTS",
  "LOAD-TYPE",
  "CONS",
  "<0?",
  "MODULO",
  "ASH",			/* 20 */
  "ROT",
  "STORE-BP-I",
  "LOAD-BP-I",
  "RETURN",
  "ALLOCATE",
  "ASSQ",
  "LOAD-LENGTH",
  "PEEK",
  "POKE",
  "MAKE-CELL",			/* 30 */
  "SUBTRACT",
  "=",
  "<",
  "BIT-NOT",
  "LONG-BRANCH",
  "LONG-BRANCH-NIL",
  "LONG-BRANCH-T",
  "LOCATE-BP-I",
  "LOAD-GLO-CON",
  "CAR",			/* 40 */
  "CDR",
  "SET-CAR",
  "SET-CDR",
  "LOCATE-CAR",
  "LOCATE-CDR",
  "PUSH-CXT-LONG",
  "CALL-PRIMITIVE",
  "THROW",
  "OBJECT-HASH",
  "OBJECT-UNHASH",		/* 50 */
  "GC",
  "BIG-ENDIAN?",
  "VLEN-ALLOCATE",
  "INC-LOC",
  "FILL-CONTINUATION",
  "CONTINUE",
  "REVERSE-CONS",
  "MOST-NEGATIVE-FIXNUM?",
  "FX-PLUS",
  "FX-TIMES",			/* 60 */
  "GET-TIME",
  "REMAINDER",
  "QUOTIENTM",
  "FULL-GC",
  "MAKE-LAMBDA",
  "GET-ARGLINE-CHAR",
  "ENABLE-ALARMS",
  "DISABLE-ALARMS",
  "RESET-ALARM-COUNTER",
  "MAKE-HEAVYWEIGHT-THREAD",	/* 70 */
  "TEST-AND-SET-LOCATIVE",
  "ILLEGAL-ARGLESS-72",
  "ILLEGAL-ARGLESS-73",
  "ILLEGAL-ARGLESS-74",
  "ILLEGAL-ARGLESS-75",
  "ILLEGAL-ARGLESS-76",
  "ILLEGAL-ARGLESS-77",
  "ILLEGAL-ARGLESS-78",
  "ILLEGAL-ARGLESS-79",
  "ILLEGAL-ARGLESS-80",		/* 80 */
  "ILLEGAL-ARGLESS-81",
  "ILLEGAL-ARGLESS-82",
  "ILLEGAL-ARGLESS-83",
  "ILLEGAL-ARGLESS-84",
  "ILLEGAL-ARGLESS-85",
  "ILLEGAL-ARGLESS-86",
  "ILLEGAL-ARGLESS-87",
  "ILLEGAL-ARGLESS-88",
  "ILLEGAL-ARGLESS-89",
  "ILLEGAL-ARGLESS-90",		/* 90 */
  "ILLEGAL-ARGLESS-91",
  "ILLEGAL-ARGLESS-92",
  "ILLEGAL-ARGLESS-93",
  "ILLEGAL-ARGLESS-94",
  "ILLEGAL-ARGLESS-95",
  "ILLEGAL-ARGLESS-96",
  "ILLEGAL-ARGLESS-97",
  "ILLEGAL-ARGLESS-98",
  "ILLEGAL-ARGLESS-99",
  "ILLEGAL-ARGLESS-100",	/* 100 */
  "ILLEGAL-ARGLESS-101",
  "ILLEGAL-ARGLESS-102",
  "ILLEGAL-ARGLESS-103",
  "ILLEGAL-ARGLESS-104",
  "ILLEGAL-ARGLESS-105",
  "ILLEGAL-ARGLESS-106",
  "ILLEGAL-ARGLESS-107",
  "ILLEGAL-ARGLESS-108",
  "ILLEGAL-ARGLESS-109",
  "ILLEGAL-ARGLESS-110",	/* 110 */
  "ILLEGAL-ARGLESS-111",
  "ILLEGAL-ARGLESS-112",
  "ILLEGAL-ARGLESS-113",
  "ILLEGAL-ARGLESS-114",
  "ILLEGAL-ARGLESS-115",
  "ILLEGAL-ARGLESS-116",
  "ILLEGAL-ARGLESS-117",
  "ILLEGAL-ARGLESS-118",
  "ILLEGAL-ARGLESS-119",
  "ILLEGAL-ARGLESS-120",	/* 120 */
  "ILLEGAL-ARGLESS-121",
  "ILLEGAL-ARGLESS-122",
  "ILLEGAL-ARGLESS-123",
  "ILLEGAL-ARGLESS-124",
  "ILLEGAL-ARGLESS-125",
  "ILLEGAL-ARGLESS-126",
  "ILLEGAL-ARGLESS-127",
};

char *instr_name[] = {
  "ILLEGAL-ARGED-0",		/* 0 */
  "HALT",
  "LOG-OP",
  "BLT-STK",
  "BRANCH-NIL",
  "BRANCH-T",
  "BRANCH",
  "POP",
  "SWAP",
  "BLAST",
  "LOAD-IMM-FIX",		/* 10 */
  "STORE-STK",
  "LOAD-BP",
  "STORE-BP",
  "LOAD-ENV",
  "STORE-ENV",
  "LOAD-STK",
  "MAKE-BP-LOC",
  "MAKE-ENV-LOC",
  "STORE-REG",
  "LOAD-REG",			/* 20 */
  "FUNCALL-CXT",
  "FUNCALL-TAIL",
  "STORE-NARGS",
  "CHECK-NARGS",
  "CHECK-NARGS-GTE",
  "STORE-SLOT",
  "LOAD-SLOT",
  "MAKE-CLOSED-ENVIRONMENT",
  "PUSH-CXT",
  "LOCATE-SLOT",		/* 30 */
  "STREAM-PRIMITIVE",
  "FILLTAG",
  "^SUPER-CXT",
  "^SUPER-TAIL",
};
